package com.yobotics.exampleSimulations.flyballGovernor;

import javax.vecmath.Vector3d;

import com.yobotics.simulationconstructionset.CylinderJoint;
import com.yobotics.simulationconstructionset.DoubleYoVariable;
import com.yobotics.simulationconstructionset.ExternalForcePoint;
import com.yobotics.simulationconstructionset.Joint;
import com.yobotics.simulationconstructionset.Link;
import com.yobotics.simulationconstructionset.LinkGraphics;
import com.yobotics.simulationconstructionset.PinJoint;
import com.yobotics.simulationconstructionset.Robot;
import com.yobotics.simulationconstructionset.YoAppearance;
import com.yobotics.simulationconstructionset.YoVariableRegistry;
import com.yobotics.simulationconstructionset.robotController.RobotController;

public class FlyballGovernorRobot extends Robot implements RobotController
{
  
   private static final long serialVersionUID = -2657468625455223170L;

   private static final double
      L1 = 0.30, R1 = 0.01, M1 = 0.1, DAMPING1 = 0.0004;
   private static final double
      L2 = 0.2, R2 = 0.005, M2 = 0.5, SPHERE_R = 0.03, DAMPING2 = 0.5;
   private static final double
      L3 = 0.1, R3 = 0.005, M3 = 0.1;
   private static final double
      L4 = 0.06, R4 = 0.03, M4 = 0.2;

   private static final double
      Ixx1 = 0.5 * M1 * L1 * L1, Iyy1 = 0.5 * M1 * L1 * L1, Izz1 = 0.5 * M1 * R1 * R1;
   private static final double
      Ixx2 = 0.5 * M2 * SPHERE_R * SPHERE_R, Iyy2 = 0.5 * M2 * SPHERE_R * SPHERE_R, Izz2 = 0.5 * M2 * SPHERE_R * SPHERE_R;
   private static final double
      Ixx3 = 0.5 * M3 * L3 * L3, Iyy3 = 0.5 * M3 * L3 * L3, Izz3 = 0.5 * M3 * R3 * R3;
   private static final double
      Ixx4 = 0.5 * M4 * L4 * L4, Iyy4 = 0.5 * M4 * L4 * L4, Izz4 = 0.5 * M4 * R4 * R4;

   public FlyballGovernorRobot()
   {
      super("FlyballGovernor");

      PinJoint rotation = new PinJoint("rotation", new Vector3d(), this, Joint.Z);
      rotation.setDamping(DAMPING1);
      Link centerRod = centerRod();
      rotation.setLink(centerRod);
      this.addRootJoint(rotation);

      // One of the flyballs

      PinJoint upperPivot1 = new PinJoint("upperPivot1", new Vector3d(R1, 0.0, L1), this, Joint.Y);
      upperPivot1.setInitialState(-0.3, 0.0);
      upperPivot1.setDamping(DAMPING2);
      upperPivot1.setLimitStops(-Math.PI / 2.0, -0.2, 100, 10);
      Link flyBall1 = flyBallLink();
      upperPivot1.setLink(flyBall1);
      rotation.addJoint(upperPivot1);

      PinJoint lowerPivot1 = new PinJoint("lowerPivot1", new Vector3d(0.0, 0.0, -2.0 / 3.0 * L2), this, Joint.Y);
      lowerPivot1.setInitialState(0.5, 0.0);
      Link loopLink1 = loopLink();
      lowerPivot1.setLink(loopLink1);
      upperPivot1.addJoint(lowerPivot1);

      ExternalForcePoint constraint1A = new ExternalForcePoint("constraint1A", new Vector3d(0.0, 0.0, -L3), this);
      lowerPivot1.addExternalForcePoint(constraint1A);

      // The other flyball

      PinJoint upperPivot2 = new PinJoint("upperPivot2", new Vector3d(-R1, 0.0, L1), this, Joint.Y);
      upperPivot2.setInitialState(0.3, 0.0);
      upperPivot2.setDamping(DAMPING2);
      upperPivot2.setLimitStops(0.2, Math.PI / 2.0, 100, 10);
      Link flyBall2 = flyBallLink();
      upperPivot2.setLink(flyBall2);
      rotation.addJoint(upperPivot2);

      PinJoint lowerPivot2 = new PinJoint("lowerPivot2", new Vector3d(0.0, 0.0, -2.0 / 3.0 * L2), this, Joint.Y);
      lowerPivot2.setInitialState(-0.5, 0.0);
      Link loopLink2 = loopLink();
      lowerPivot2.setLink(loopLink2);
      upperPivot2.addJoint(lowerPivot2);

      ExternalForcePoint constraint2A = new ExternalForcePoint("constraint2A", new Vector3d(0.0, 0.0, -L3), this);
      lowerPivot2.addExternalForcePoint(constraint2A);


      // The sliding Cylinder:

      CylinderJoint cylinderJoint = new CylinderJoint("cylinder_theta", "cylinder_z", new Vector3d(), this, Joint.Z);
      cylinderJoint.setInitialState(0.0, 0.0, 0.05, 0.0);
      Link cylinderLink = cylinderLink();
      cylinderJoint.setLink(cylinderLink);
      this.addRootJoint(cylinderJoint);

      ExternalForcePoint constraint1B = new ExternalForcePoint("constraint1B", new Vector3d(R4, 0.0, L4 / 2.0), this);
      ExternalForcePoint constraint2B = new ExternalForcePoint("constraint2B", new Vector3d(-R4, 0.0, L4 / 2.0), this);
      cylinderJoint.addExternalForcePoint(constraint1B);
      cylinderJoint.addExternalForcePoint(constraint2B);


      this.setController(this);
      this.initControl();

   }



   private Link centerRod()
   {
      Link ret = new Link("Center Rod");
      ret.setMass(M1);
      ret.setMomentOfInertia(0.0, 0.0, Izz1);
      ret.setComOffset(0.0, 0.0, L1 / 2.0);
      
      LinkGraphics linkGraphics = new LinkGraphics();
      linkGraphics.addCylinder(L1, R1, YoAppearance.Red());
      ret.setLinkGraphics(linkGraphics);
      
      return ret;
   }

   private Link flyBallLink()
   {
      Link ret = new Link("Flyball");
      ret.setMass(M2);
      ret.setMomentOfInertia(Ixx2, Iyy2, Izz2);
      ret.setComOffset(0.0, 0.0, -L2);
      
      LinkGraphics linkGraphics = new LinkGraphics();
      linkGraphics.translate(0.0, 0.0, -L2);
      linkGraphics.addCylinder(L2, R2);
      linkGraphics.addSphere(SPHERE_R, YoAppearance.DarkGreen());
      ret.setLinkGraphics(linkGraphics);
      
      return ret;
   }


   private Link loopLink()
   {
      Link ret = new Link("Loop");
      ret.setMass(M3);
      ret.setMomentOfInertia(Ixx3, Iyy3, Izz3);
      ret.setComOffset(0.0, 0.0, -L3 / 2.0);

      LinkGraphics linkGraphics = new LinkGraphics();
      linkGraphics.translate(0.0, 0.0, -L3);
      linkGraphics.addCylinder(L3, R3);
      ret.setLinkGraphics(linkGraphics);

      return ret;
   }

   private Link cylinderLink()
   {
      Link ret = new Link("Cylinder Link");
      ret.setMass(M4);
      ret.setMomentOfInertia(Ixx4, Iyy4, Izz4);
      ret.setComOffset(0.0, 0.0, L4 / 2.0);

      LinkGraphics linkGraphics = new LinkGraphics();
      linkGraphics.addCylinder(L4, R4, YoAppearance.DarkBlue());
      linkGraphics.addCylinder(L4 / 8.0, 1.1 * R4);
      linkGraphics.translate(0.0, 0.0, 7.0 / 8.0 * L4);
      linkGraphics.addCylinder(L4 / 8.0, 1.1 * R4);
      ret.setLinkGraphics(linkGraphics);

      return ret;
   }

   private DoubleYoVariable constraint1A_x, constraint1A_y, constraint1A_z, constraint1A_dx, constraint1A_dy, constraint1A_dz;
   private DoubleYoVariable constraint1A_fx, constraint1A_fy, constraint1A_fz;
   private DoubleYoVariable constraint2A_x, constraint2A_y, constraint2A_z, constraint2A_dx, constraint2A_dy, constraint2A_dz;
   private DoubleYoVariable constraint2A_fx, constraint2A_fy, constraint2A_fz;

   private DoubleYoVariable constraint1B_x, constraint1B_y, constraint1B_z, constraint1B_dx, constraint1B_dy, constraint1B_dz;
   private DoubleYoVariable constraint1B_fx, constraint1B_fy, constraint1B_fz;
   private DoubleYoVariable constraint2B_x, constraint2B_y, constraint2B_z, constraint2B_dx, constraint2B_dy, constraint2B_dz;
   private DoubleYoVariable constraint2B_fx, constraint2B_fy, constraint2B_fz;

   private DoubleYoVariable tau_rotation, q_cylinder_z, qd_cylinder_z;

   private final YoVariableRegistry registry = new YoVariableRegistry("FlyballGovernorController");

   private DoubleYoVariable k_feedback = new DoubleYoVariable("k_feedback", registry);
   private DoubleYoVariable q_d_cylinder_z = new DoubleYoVariable("q_d_cylinder_z", registry);

   public DoubleYoVariable[] getControlVars()
   {
      return new DoubleYoVariable[] {k_feedback, q_d_cylinder_z};
   }

   public void initControl()
   {
      constraint1A_x = (DoubleYoVariable)this.getVariable("constraint1A_x");
      constraint1A_y = (DoubleYoVariable)this.getVariable("constraint1A_y");
      constraint1A_z = (DoubleYoVariable)this.getVariable("constraint1A_z");
      constraint1A_dx = (DoubleYoVariable)this.getVariable("constraint1A_dx");
      constraint1A_dy = (DoubleYoVariable)this.getVariable("constraint1A_dy");
      constraint1A_dz = (DoubleYoVariable)this.getVariable("constraint1A_dz");
      constraint1A_fx = (DoubleYoVariable)this.getVariable("constraint1A_fx");
      constraint1A_fy = (DoubleYoVariable)this.getVariable("constraint1A_fy");
      constraint1A_fz = (DoubleYoVariable)this.getVariable("constraint1A_fz");

      constraint2A_x = (DoubleYoVariable)this.getVariable("constraint2A_x");
      constraint2A_y = (DoubleYoVariable)this.getVariable("constraint2A_y");
      constraint2A_z = (DoubleYoVariable)this.getVariable("constraint2A_z");
      constraint2A_dx = (DoubleYoVariable)this.getVariable("constraint2A_dx");
      constraint2A_dy = (DoubleYoVariable)this.getVariable("constraint2A_dy");
      constraint2A_dz = (DoubleYoVariable)this.getVariable("constraint2A_dz");
      constraint2A_fx = (DoubleYoVariable)this.getVariable("constraint2A_fx");
      constraint2A_fy = (DoubleYoVariable)this.getVariable("constraint2A_fy");
      constraint2A_fz = (DoubleYoVariable)this.getVariable("constraint2A_fz");

      constraint1B_x = (DoubleYoVariable)this.getVariable("constraint1B_x");
      constraint1B_y = (DoubleYoVariable)this.getVariable("constraint1B_y");
      constraint1B_z = (DoubleYoVariable)this.getVariable("constraint1B_z");
      constraint1B_dx = (DoubleYoVariable)this.getVariable("constraint1B_dx");
      constraint1B_dy = (DoubleYoVariable)this.getVariable("constraint1B_dy");
      constraint1B_dz = (DoubleYoVariable)this.getVariable("constraint1B_dz");
      constraint1B_fx = (DoubleYoVariable)this.getVariable("constraint1B_fx");
      constraint1B_fy = (DoubleYoVariable)this.getVariable("constraint1B_fy");
      constraint1B_fz = (DoubleYoVariable)this.getVariable("constraint1B_fz");

      constraint2B_x = (DoubleYoVariable)this.getVariable("constraint2B_x");
      constraint2B_y = (DoubleYoVariable)this.getVariable("constraint2B_y");
      constraint2B_z = (DoubleYoVariable)this.getVariable("constraint2B_z");
      constraint2B_dx = (DoubleYoVariable)this.getVariable("constraint2B_dx");
      constraint2B_dy = (DoubleYoVariable)this.getVariable("constraint2B_dy");
      constraint2B_dz = (DoubleYoVariable)this.getVariable("constraint2B_dz");
      constraint2B_fx = (DoubleYoVariable)this.getVariable("constraint2B_fx");
      constraint2B_fy = (DoubleYoVariable)this.getVariable("constraint2B_fy");
      constraint2B_fz = (DoubleYoVariable)this.getVariable("constraint2B_fz");

      tau_rotation = (DoubleYoVariable)this.getVariable("tau_rotation");
      q_cylinder_z = (DoubleYoVariable)this.getVariable("q_cylinder_z");
      qd_cylinder_z = (DoubleYoVariable)this.getVariable("qd_cylinder_z");

      k_feedback.set(0.2);
      q_d_cylinder_z.set(0.1);
   }

   public void doControl()
   {
      double K = 10000.0, B = 15.0;

      constraint1A_fx.set(K * (constraint1B_x.getDoubleValue() - constraint1A_x.getDoubleValue()) + B * (constraint1B_dx.getDoubleValue() - constraint1A_dx.getDoubleValue()));
      constraint1A_fy.set(K * (constraint1B_y.getDoubleValue() - constraint1A_y.getDoubleValue()) + B * (constraint1B_dy.getDoubleValue() - constraint1A_dy.getDoubleValue()));
      constraint1A_fz.set(K * (constraint1B_z.getDoubleValue() - constraint1A_z.getDoubleValue()) + B * (constraint1B_dz.getDoubleValue() - constraint1A_dz.getDoubleValue()));

      constraint1B_fx.set(-constraint1A_fx.getDoubleValue());
      constraint1B_fy.set(-constraint1A_fy.getDoubleValue());
      constraint1B_fz.set(-constraint1A_fz.getDoubleValue());

      constraint2A_fx.set(K * (constraint2B_x.getDoubleValue() - constraint2A_x.getDoubleValue()) + B * (constraint2B_dx.getDoubleValue() - constraint2A_dx.getDoubleValue()));
      constraint2A_fy.set(K * (constraint2B_y.getDoubleValue() - constraint2A_y.getDoubleValue()) + B * (constraint2B_dy.getDoubleValue() - constraint2A_dy.getDoubleValue()));
      constraint2A_fz.set(K * (constraint2B_z.getDoubleValue() - constraint2A_z.getDoubleValue()) + B * (constraint2B_dz.getDoubleValue() - constraint2A_dz.getDoubleValue()));

      constraint2B_fx.set(-constraint2A_fx.getDoubleValue());
      constraint2B_fy.set(-constraint2A_fy.getDoubleValue());
      constraint2B_fz.set(-constraint2A_fz.getDoubleValue());


      // Feedback:

      tau_rotation.set(k_feedback.getDoubleValue() * (q_d_cylinder_z.getDoubleValue() - q_cylinder_z.getDoubleValue()));
   }

   public YoVariableRegistry getYoVariableRegistry()
   {
      return registry;
   }
   
   public void initialize()
   {      
   }

   public String getDescription()
   {
      return getName();
   }


}
