package com.yobotics.exampleSimulations.springflamingo;

import java.awt.Container;

import javax.swing.BoxLayout;
import javax.swing.JFrame;

import com.yobotics.simulationconstructionset.DoubleYoVariable;
import com.yobotics.simulationconstructionset.YoVariableRegistry;
import com.yobotics.simulationconstructionset.robotController.RobotController;
import com.yobotics.simulationconstructionset.util.statemachines.State;
import com.yobotics.simulationconstructionset.util.statemachines.StateMachine;
import com.yobotics.simulationconstructionset.util.statemachines.StateMachinesJPanel;
import com.yobotics.simulationconstructionset.util.statemachines.StateTransition;
import com.yobotics.simulationconstructionset.util.statemachines.StateTransitionCondition;

/**
 * <p>Title: SpringFlamingoController</p>
 *
 * <p>Description: Simple balistic walking controller for the SpringFlamingo simulation model.
 * Controls body pitch through hip torque, velocity through ankle torque. </p>
 *
 * <p>Copyright: Copyright (c) 2001</p>
 *
 * <p>Company: Yobotics, Inc.</p>
 *
 * @author not attributable
 * @version 1.0
 */
public class SpringFlamingoController implements RobotController
{
   private static final long serialVersionUID = 8632702461366092963L;

   private final YoVariableRegistry registry = new YoVariableRegistry("SpringFlamingoController");

   // State Machine:
   private enum States {SUPPORT, TOE_OFF, SWING, STRAIGHTEN}

   ;
   private StateMachine leftStateMachine, rightStateMachine;

   // Control Parameters:
   private final DoubleYoVariable stand_gain = new DoubleYoVariable("stand_gain", "Gain for torquing the support ankle based on an error in desired x",
                                                  registry);
   private final DoubleYoVariable x_d = new DoubleYoVariable("x_d", "Desired x location. Controlled with stand_gain", registry);

   private final DoubleYoVariable v_nom = new DoubleYoVariable("v_nom", "Nominal velocity", registry);
   private final DoubleYoVariable vtp_gain = new DoubleYoVariable("vtp_gain", "Gain on velocity error to support ankle torque.", registry);
   private final DoubleYoVariable t_d = new DoubleYoVariable("t_d", "Desired body pitch", registry);
   private final DoubleYoVariable t_gain = new DoubleYoVariable("t_gain", "Gain on Pitch angle to hip torque", registry);
   private final DoubleYoVariable t_damp = new DoubleYoVariable("t_damp", "Hip damping", registry);

   private final DoubleYoVariable
      knee_d = new DoubleYoVariable("knee_d", registry), knee_gain = new DoubleYoVariable("knee_gain", registry),
      knee_damp = new DoubleYoVariable("knee_damp", registry);
   private final DoubleYoVariable
      hip_d = new DoubleYoVariable("hip_d", registry), hip_gain = new DoubleYoVariable("hip_gain", registry),
      hip_damp = new DoubleYoVariable("hip_damp", registry), hip_hold = new DoubleYoVariable("hip_hold", registry);
   private final DoubleYoVariable
      swing_gain_knee = new DoubleYoVariable("swing_gain_knee", registry), swing_damp_knee = new DoubleYoVariable("swing_damp_knee", registry),
      force_thresh = new DoubleYoVariable("force_thresh", registry);

   private final DoubleYoVariable min_support_time = new DoubleYoVariable("min_support_time", registry);
   private final DoubleYoVariable
      swing_time = new DoubleYoVariable("swing_time", registry), toe_off_time = new DoubleYoVariable("toe_off_time", registry);
   private final DoubleYoVariable ankle_d = new DoubleYoVariable("ankle_d", "Swing ankle desired pitch", registry);
   private final DoubleYoVariable ankle_gain = new DoubleYoVariable("ankle_gain", "Gain on ankle ankle to ankle torque", registry);
   private final DoubleYoVariable ankle_damp = new DoubleYoVariable("ankle_damp", "Gain on ankle velocity to ankle torque", registry);

   private final DoubleYoVariable ankle_limit_set = new DoubleYoVariable("ankle_limit_set", registry);
   private final DoubleYoVariable ankle_limit_gain = new DoubleYoVariable("ankle_limit_gain", registry);

// private final YoVariable toe_off_ankle_thresh = new YoVariable("toe_off_ankle_thresh", registry);
// private final YoVariable ankle_push = new YoVariable("ankle_push", registry);

   private final DoubleYoVariable push_set = new DoubleYoVariable("push_set", "Ankle angle to toe off to", registry);
   private final DoubleYoVariable push_gain = new DoubleYoVariable("push_gain", "Gain on ankle angle to ankle torque when toeing off", registry);
   private final DoubleYoVariable push_damp = new DoubleYoVariable("push_damp", "Damping on ankle angle to ankle torque when toeing off", registry);

   private final DoubleYoVariable knee_collapse = new DoubleYoVariable("knee_collapse", registry);

   // Active and Passive Torques:
   private final DoubleYoVariable
      act_lh = new DoubleYoVariable("act_lh", registry), act_lk = new DoubleYoVariable("act_lk", registry), act_la = new DoubleYoVariable("act_la", registry),
      pas_la = new DoubleYoVariable("pas_la", registry);
   private final DoubleYoVariable
      act_rh = new DoubleYoVariable("act_rh", registry), act_rk = new DoubleYoVariable("act_rk", registry), act_ra = new DoubleYoVariable("act_ra", registry),
      pas_ra = new DoubleYoVariable("pas_ra", registry);
   private final DoubleYoVariable
      pas_lh = new DoubleYoVariable("pas_lh", registry), pas_lk = new DoubleYoVariable("pas_lk", registry), pas_rh = new DoubleYoVariable("pas_rh", registry),
      pas_rk = new DoubleYoVariable("pas_rk", registry);
   private final DoubleYoVariable max_hip_torque = new DoubleYoVariable("max_hip_torque", registry);

   // Power and Energy:
   private final DoubleYoVariable
      pow_lh = new DoubleYoVariable("pow_lh", registry), pow_lk = new DoubleYoVariable("pow_lk", registry), pow_la = new DoubleYoVariable("pow_la", registry);
   private final DoubleYoVariable
      pow_rh = new DoubleYoVariable("pow_rh", registry), pow_rk = new DoubleYoVariable("pow_rk", registry), pow_ra = new DoubleYoVariable("pow_ra", registry);
   private final DoubleYoVariable
      power = new DoubleYoVariable("power", registry), energy = new DoubleYoVariable("energy", registry);

   // Computed State Variables:
   private final DoubleYoVariable vel = new DoubleYoVariable("vel", "Actual Velocity", registry);
   private final DoubleYoVariable left_heel = new DoubleYoVariable("left_heel", "Orientation of the left foot", registry);
   private final DoubleYoVariable right_heel = new DoubleYoVariable("right_heel", "Orientation of the right foot", registry);

   private final DoubleYoVariable left_force = new DoubleYoVariable("left_force", "Total vertical ground reaction force on the left foot", registry);
   private final DoubleYoVariable left_cop = new DoubleYoVariable("left_cop", "Location of the left foots Center of Pressure", registry);
   private final DoubleYoVariable right_force = new DoubleYoVariable("right_force", "Total vertical ground reaction force on the left foot", registry);
   private final DoubleYoVariable right_cop = new DoubleYoVariable("right_cop", "Location of the right foots Center of Pressure", registry);

   private final DoubleYoVariable left_hip_set = new DoubleYoVariable("left_hip_set", registry);
   private final DoubleYoVariable right_hip_set = new DoubleYoVariable("right_hip_set", registry);


   private final SpringFlamingoRobot robot;

   private String name;

   public SpringFlamingoController(SpringFlamingoRobot robot, String name)
   {
      this.name = name;
      this.robot = robot;
      initControl();
      setupStateMachines();

      createStateMachineWindow();
   }

   public void createStateMachineWindow()
   {
      JFrame jFrame = new JFrame("Spring Flamingo State Machines");
      Container contentPane = jFrame.getContentPane();
      contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.X_AXIS));
      
      StateMachinesJPanel leftStateMachinePanel = new StateMachinesJPanel(leftStateMachine, false);
      StateMachinesJPanel rightStateMachinePanel = new StateMachinesJPanel(rightStateMachine, true);
      
      jFrame.getContentPane().add(leftStateMachinePanel);
      jFrame.getContentPane().add(rightStateMachinePanel);
      
      jFrame.pack();
      jFrame.setSize(450, 300);
      jFrame.setAlwaysOnTop(true);
      jFrame.setVisible(true);

      // Doing the following will cause redraw when the state changes, but not during replay or rewind:
      leftStateMachine.attachStateChangedListener(leftStateMachinePanel);
      rightStateMachine.attachStateChangedListener(rightStateMachinePanel);

      // Doing this will cause redraw every specified miliseconds:
//    leftStateMachinePanel.createUpdaterThread(250);
//    rightStateMachinePanel.createUpdaterThread(250);
   }

   public YoVariableRegistry getYoVariableRegistry()
   {
      return registry;
   }

   private void initControl()
   {
      // Initialize the variables.
      robot.initializeForBallisticWalking();

//    stand_gain.val = 0;
//    x_d.val = 0;
//    v_nom.val = -0.4;
//    vtp_gain.val = 10.0; //0.0;
//    t_d.val = 0;
//    t_gain.val = 100;
//    t_damp.val = 20;
//    knee_d.val = 0;
//    knee_gain.val = 30;
//    knee_damp.val = 10;
//    hip_d.val = 0.587059;
//    hip_gain.val = 15.9804;
//    hip_damp.val = 2.01177;
//    hip_hold.val = 0.358627;
//    swing_gain_knee.val = 1.81176;
//    swing_damp_knee.val = 0.403529;
//    force_thresh.val = 13.6078;
//    min_support_time.val = 0.2;
//    swing_time.val = 0.418039;
//    toe_off_time.val = 0.05;
//    ankle_d.val = 0;
//    ankle_gain.val = 4;
//    ankle_damp.val = 1;
//    ankle_limit_set.val = 0;
//    ankle_limit_gain.val = 250.98;
//    left_force.val = 0;
//    left_cop.val = 0.5;
//    right_force.val = 131.83;
//    right_cop.val = 0.557913;
//    left_hip_set.val = 0.358627;
//    left_heel.val = -0.392589;
//    right_hip_set.val = 0.358627;
//    right_heel.val = -0.231515;
////    toe_off_ankle_thresh.val = 0.1;
////    ankle_push.val = -0.3;
//    push_gain.val = 9.56078;
//    knee_collapse.val = 0;
//    push_set.val = -0.236471;
//    push_damp.val = 0;
//
//    robot.t.val = 0.0; //10.00;
//
//    robot.q_x.val = -6.68245; robot.qd_x.val = -0.745821;
//    robot.q_z.val = 0.859601; robot.qd_z.val = -0.127678;
//    robot.q_pitch.val = 0.0158429; robot.qd_pitch.val = 0.293524;
//    robot.q_rh.val = -0.210433; robot.qd_rh.val = -1.10499;
//    robot.q_rk.val = -1.43749e-05; robot.qd_rk.val = 0.0279385;
//    robot.q_ra.val = 0.19298; robot.qd_ra.val = 0.685901;
//    robot.q_lh.val = 0.443428; robot.qd_lh.val = 0.727701;
//    robot.q_lk.val = -0.100803; robot.qd_lk.val = -2.09991;
//    robot.q_la.val = 0.0329076; robot.qd_la.val = -1.93702;
//


      vel.set(0.0);
      stand_gain.set(0.0);
      x_d.set(0.0);
      v_nom.set(-0.4);
      vtp_gain.set(0.0);
      t_d.set(0.0);
      t_gain.set(100.0);
      t_damp.set(20.0);
      knee_d.set(0.0);
      knee_gain.set(30.0);
      knee_damp.set(10.0);
      hip_d.set(0.587059);
      hip_gain.set(15.9804);
      hip_damp.set(2.01177);
      hip_hold.set(0.358627);
      swing_gain_knee.set(1.81176);
      swing_damp_knee.set(0.403529);
      force_thresh.set(13.6078);
      min_support_time.set(0.2);
      swing_time.set(0.418039);
      toe_off_time.set(0.05);
      ankle_d.set(0.0);
      ankle_gain.set(4.0);
      ankle_damp.set(1.0);
      ankle_limit_set.set(0.0);
      ankle_limit_gain.set(250.98);

//    left_state.val = 3.0;
//    right_state.val = 0.0;
      left_force.set(0.0);
      left_cop.set(0.5);
      right_force.set(131.83);
      right_cop.set(0.557913);
      left_hip_set.set(0.358627);

//    left_switch_time.val = 0.0;
      left_heel.set(-0.392589);
      right_hip_set.set(0.358627);

//    right_switch_time.val = 0.0;
      right_heel.set(-0.231515);
      push_gain.set(9.56078);
      knee_collapse.set(0.0);
      push_set.set(-0.236471);
      push_damp.set(0.0);
      pow_lh.set(0.0);
      pow_lk.set(0.0);
      pow_la.set(0.0);
      pow_rh.set(0.0);
      pow_rk.set(0.0);
      pow_ra.set(0.0);
      power.set(0.0);
      energy.set(0.0);
      act_lh.set(0.0);
      act_lk.set(0.0);
      act_la.set(0.0);
      pas_la.set(0.0);
      act_rh.set(0.0);
      act_rk.set(0.0);
      act_ra.set(0.0);
      pas_ra.set(0.0);
      pas_lh.set(0.0);
      pas_lk.set(0.0);
      pas_rh.set(0.0);
      pas_rk.set(0.0);
      max_hip_torque.set(0.0);

   }

   public void doControl()
   {
      balistic_walking_state_machine();
   }

   public double passive_ankle_torques(double pos, double vel)
   {
      if (pos > ankle_limit_set.getDoubleValue())
         return (-ankle_limit_gain.getDoubleValue() * (ankle_limit_set.getDoubleValue() - pos) * (ankle_limit_set.getDoubleValue() - pos));
      else
         return (0.0);
   }


   private double toe_off_ankle_torques(double pos, double vel)
   {
      return (push_gain.getDoubleValue() * (push_set.getDoubleValue() - pos) - push_damp.getDoubleValue() * vel);
   }

   private void setupStateMachines()
   {
      // States and Actions:

      State leftSupportState = new State(States.SUPPORT)
      {
         public void doAction()
         {
            // Use hip to servo pitch
            if (left_force.getDoubleValue() > 20.0)
               act_lh.set(-t_gain.getDoubleValue() * (t_d.getDoubleValue() - robot.q_pitch.getDoubleValue())
                          + t_damp.getDoubleValue() * robot.qd_pitch.getDoubleValue());

            // Keep knee straight
            act_lk.set(knee_gain.getDoubleValue() * (knee_d.getDoubleValue() - robot.q_lk.getDoubleValue())
                       - knee_damp.getDoubleValue() * robot.qd_lk.getDoubleValue());

            // Use ankle to servo speed, position
            act_la.set(-stand_gain.getDoubleValue() * (x_d.getDoubleValue() - robot.q_x.getDoubleValue())
                       - vtp_gain.getDoubleValue() * (v_nom.getDoubleValue() - robot.qd_x.getDoubleValue()));

            // Ankle limit to go on toes and maintain cop
            pas_la.set(passive_ankle_torques(robot.q_la.getDoubleValue(), robot.qd_la.getDoubleValue()));

         }

         public void doTransitionIntoAction()
         {
         }
         public void doTransitionOutOfAction()
         {
         }
      };


      State leftToeOffState = new State(States.TOE_OFF)
      {
         public void doAction()
         {
            // Use hip to servo pitch
            act_lh.set(-t_gain.getDoubleValue() * (t_d.getDoubleValue() - robot.q_pitch.getDoubleValue())
                       + t_damp.getDoubleValue() * robot.qd_pitch.getDoubleValue());

            // Keep knee straight
            act_lk.set(knee_gain.getDoubleValue() * (knee_d.getDoubleValue() - robot.q_lk.getDoubleValue())
                       - knee_damp.getDoubleValue() * robot.qd_lk.getDoubleValue());

            // Use ankle to servo speed, position
            act_la.set(-stand_gain.getDoubleValue() * (x_d.getDoubleValue() - robot.q_x.getDoubleValue())
                       - vtp_gain.getDoubleValue() * (v_nom.getDoubleValue() - robot.qd_x.getDoubleValue()));

            // Ankle limit to go on toes and maintain cop
            pas_la.set(passive_ankle_torques(robot.q_la.getDoubleValue(), robot.qd_la.getDoubleValue()));

            // Ankle push off
            act_la.set(act_la.getDoubleValue() + toe_off_ankle_torques(robot.q_la.getDoubleValue(), robot.qd_la.getDoubleValue()));

         }

         public void doTransitionIntoAction()
         {
         }
         public void doTransitionOutOfAction()
         {
         }
      };

      State leftSwingState = new State(States.SWING)
      {
         public void doAction()
         {
            // Servo hip up
            left_hip_set.set(hip_d.getDoubleValue());
            act_lh.set(hip_gain.getDoubleValue() * (left_hip_set.getDoubleValue() - robot.q_lh.getDoubleValue())
                       - hip_damp.getDoubleValue() * robot.qd_lh.getDoubleValue());

            // Damp the knee
            act_lk.set(-swing_damp_knee.getDoubleValue() * robot.qd_lk.getDoubleValue());

            // act_lk.val = 0.0;

            // Damp and Toggle the knee:
            act_lk.set(act_lk.getDoubleValue() + knee_collapse.getDoubleValue());

            // Continue toe off until toe leaves ground
            if (robot.gc_ltoe_fs.getDoubleValue() > 0.1)
            {
               pas_la.set(passive_ankle_torques(robot.q_la.getDoubleValue(), robot.qd_la.getDoubleValue()));
               act_la.set(toe_off_ankle_torques(robot.q_la.getDoubleValue(), robot.qd_la.getDoubleValue()));
            }
            else
            {
               // Servo ankle level to the ground
               left_heel.set(-robot.q_pitch.getDoubleValue() - robot.q_lh.getDoubleValue() - robot.q_lk.getDoubleValue() - robot.q_la.getDoubleValue());
               act_la.set(-ankle_gain.getDoubleValue() * (ankle_d.getDoubleValue() - left_heel.getDoubleValue())
                          - ankle_damp.getDoubleValue() * robot.qd_la.getDoubleValue());
            }

         }

         public void doTransitionIntoAction()
         {
         }
         public void doTransitionOutOfAction()
         {
         }
      };

      State leftStraightenState = new State(States.STRAIGHTEN)
      {
         public void doAction()
         {
            // Servo hip to a more shallow angle
            left_hip_set.set(hip_hold.getDoubleValue());
            act_lh.set(hip_gain.getDoubleValue() * (left_hip_set.getDoubleValue() - robot.q_lh.getDoubleValue())
                       - hip_damp.getDoubleValue() * robot.qd_lh.getDoubleValue());

            // Keep knee straight
            act_lk.set(swing_gain_knee.getDoubleValue() * (knee_d.getDoubleValue() - robot.q_lk.getDoubleValue())
                       - swing_damp_knee.getDoubleValue() * robot.qd_lk.getDoubleValue());

            // Servo ankle level to the ground
            left_heel.set(-robot.q_pitch.getDoubleValue() - robot.q_lh.getDoubleValue() - robot.q_lk.getDoubleValue() - robot.q_la.getDoubleValue());
            act_la.set(-ankle_gain.getDoubleValue() * (ankle_d.getDoubleValue() - left_heel.getDoubleValue())
                       - ankle_damp.getDoubleValue() * robot.qd_la.getDoubleValue());
         }

         public void doTransitionIntoAction()
         {
         }
         public void doTransitionOutOfAction()
         {
         }
      };


      State rightSupportState = new State(States.SUPPORT)
      {
         public void doAction()
         {
            // Use hip to servo pitch
            if (right_force.getDoubleValue() > 20.0)
               act_rh.set(-t_gain.getDoubleValue() * (t_d.getDoubleValue() - robot.q_pitch.getDoubleValue())
                          + t_damp.getDoubleValue() * robot.qd_pitch.getDoubleValue());

            // Keep knee straight
            act_rk.set(knee_gain.getDoubleValue() * (knee_d.getDoubleValue() - robot.q_rk.getDoubleValue())
                       - knee_damp.getDoubleValue() * robot.qd_rk.getDoubleValue());

            // Use ankle to servo speed, position
            act_ra.set(-stand_gain.getDoubleValue() * (x_d.getDoubleValue() - robot.q_x.getDoubleValue())
                       - vtp_gain.getDoubleValue() * (v_nom.getDoubleValue() - robot.qd_x.getDoubleValue()));

            // Ankle limit to go on toes and maintain cop
            pas_ra.set(passive_ankle_torques(robot.q_ra.getDoubleValue(), robot.qd_ra.getDoubleValue()));
         }

         public void doTransitionIntoAction()
         {
         }
         public void doTransitionOutOfAction()
         {
         }
      };

      State rightToeOffState = new State(States.TOE_OFF)
      {
         public void doAction()
         {
            // Use hip to servo pitch
            act_rh.set(-t_gain.getDoubleValue() * (t_d.getDoubleValue() - robot.q_pitch.getDoubleValue())
                       + t_damp.getDoubleValue() * robot.qd_pitch.getDoubleValue());

            // Keep knee straight
            act_rk.set(knee_gain.getDoubleValue() * (knee_d.getDoubleValue() - robot.q_rk.getDoubleValue())
                       - knee_damp.getDoubleValue() * robot.qd_rk.getDoubleValue());

            // Use ankle to servo speed, position
            act_ra.set(-stand_gain.getDoubleValue() * (x_d.getDoubleValue() - robot.q_x.getDoubleValue())
                       - vtp_gain.getDoubleValue() * (v_nom.getDoubleValue() - robot.qd_x.getDoubleValue()));

            // Ankle limit to go on toes and maintain cop
            pas_ra.set(passive_ankle_torques(robot.q_ra.getDoubleValue(), robot.qd_ra.getDoubleValue()));

            // Ankle push off */
            act_ra.set(act_ra.getDoubleValue() + toe_off_ankle_torques(robot.q_ra.getDoubleValue(), robot.qd_ra.getDoubleValue()));

         }

         public void doTransitionIntoAction()
         {
         }
         public void doTransitionOutOfAction()
         {
         }
      };

      State rightSwingState = new State(States.SWING)
      {
         public void doAction()
         {
            // Servo hip up
            right_hip_set.set(hip_d.getDoubleValue());
            act_rh.set(hip_gain.getDoubleValue() * (right_hip_set.getDoubleValue() - robot.q_rh.getDoubleValue())
                       - hip_damp.getDoubleValue() * robot.qd_rh.getDoubleValue());

            // Damp the knee
            act_rk.set(-swing_damp_knee.getDoubleValue() * robot.qd_rk.getDoubleValue());

            // Damp and Toggle the knee
            act_rk.set(act_rk.getDoubleValue() + knee_collapse.getDoubleValue());

            // Continue toe off until toe leaves ground
            if (robot.gc_rtoe_fs.getDoubleValue() > 0.1)
            {
               pas_ra.set(passive_ankle_torques(robot.q_ra.getDoubleValue(), robot.qd_ra.getDoubleValue()));
               act_ra.set(toe_off_ankle_torques(robot.q_ra.getDoubleValue(), robot.qd_ra.getDoubleValue()));
            }
            else
            {
               // Servo ankle level to the ground
               right_heel.set(-robot.q_pitch.getDoubleValue() - robot.q_rh.getDoubleValue() - robot.q_rk.getDoubleValue() - robot.q_ra.getDoubleValue());
               act_ra.set(-ankle_gain.getDoubleValue() * (ankle_d.getDoubleValue() - right_heel.getDoubleValue())
                          - ankle_damp.getDoubleValue() * robot.qd_ra.getDoubleValue());
            }
         }

         public void doTransitionIntoAction()
         {
         }
         public void doTransitionOutOfAction()
         {
         }
      };

      State rightStraightenState = new State(States.STRAIGHTEN)
      {
         public void doAction()
         {
            // Servo hip to a more shallow angle
            right_hip_set.set(hip_hold.getDoubleValue());
            act_rh.set(hip_gain.getDoubleValue() * (right_hip_set.getDoubleValue() - robot.q_rh.getDoubleValue())
                       - hip_damp.getDoubleValue() * robot.qd_rh.getDoubleValue());

            // Keep knee straight
            act_rk.set(swing_gain_knee.getDoubleValue() * (knee_d.getDoubleValue() - robot.q_rk.getDoubleValue())
                       - swing_damp_knee.getDoubleValue() * robot.qd_rk.getDoubleValue());

            // Servo ankle level to the ground
            right_heel.set(-robot.q_pitch.getDoubleValue() - robot.q_rh.getDoubleValue() - robot.q_rk.getDoubleValue() - robot.q_ra.getDoubleValue());
            act_ra.set(-ankle_gain.getDoubleValue() * (ankle_d.getDoubleValue() - right_heel.getDoubleValue())
                       - ankle_damp.getDoubleValue() * robot.qd_ra.getDoubleValue());
         }

         public void doTransitionIntoAction()
         {
         }
         public void doTransitionOutOfAction()
         {
         }
      };

      // Create the state machines:
      leftStateMachine = new StateMachine("leftState", "leftSwitchTime", States.class, robot.t, registry);
      rightStateMachine = new StateMachine("rightState", "rightSwitchTime", States.class, robot.t, registry);

      // Transition Actions:
      // StateTransitionAction resetLeftSwitchTime = new ResetSwitchTimeAction(left_switch_time);
      // StateTransitionAction resetRightSwitchTime = new ResetSwitchTimeAction(right_switch_time);

      // Transition Conditions:
      StateTransitionCondition leftHealUnloaded = new LeftHealUnloadedCondition();
      StateTransitionCondition leftFootUnloaded = new LeftFootUnloadedCondition();
      StateTransitionCondition leftFootTouchedDown = new LeftFootTouchedDownCondition();

      StateTransitionCondition rightHealUnloaded = new RightHealUnloadedCondition();
      StateTransitionCondition rightFootUnloaded = new RightFootUnloadedCondition();
      StateTransitionCondition rightFootTouchedDown = new RightFootTouchedDownCondition();

      // Left State Transitions:
      StateTransition leftSupportToToeOff = new StateTransition(States.TOE_OFF, leftHealUnloaded);
      leftSupportToToeOff.addTimePassedCondition(min_support_time);
      leftSupportState.addStateTransition(leftSupportToToeOff);

      StateTransition leftToeOffToSwing = new StateTransition(States.SWING, leftFootUnloaded);
      leftToeOffState.addStateTransition(leftToeOffToSwing);

      StateTransition leftSwingToStraighten = new StateTransition(States.STRAIGHTEN, swing_time);
      leftSwingState.addStateTransition(leftSwingToStraighten);

      StateTransition leftStraightenToSupport = new StateTransition(States.SUPPORT, leftFootTouchedDown);
      leftStraightenState.addStateTransition(leftStraightenToSupport);

      // Right State Transitions:
      StateTransition rightSupportToToeOff = new StateTransition(States.TOE_OFF, rightHealUnloaded);
      rightSupportToToeOff.addTimePassedCondition(min_support_time);
      rightSupportState.addStateTransition(rightSupportToToeOff);

      StateTransition rightToeOffToSwing = new StateTransition(States.SWING, rightFootUnloaded);
      rightToeOffState.addStateTransition(rightToeOffToSwing);

      StateTransition rightSwingToStraighten = new StateTransition(States.STRAIGHTEN, swing_time);
      rightSwingState.addStateTransition(rightSwingToStraighten);

      StateTransition rightStraightenToSupport = new StateTransition(States.SUPPORT, rightFootTouchedDown);
      rightStraightenState.addStateTransition(rightStraightenToSupport);


      // Assemble the Left State Machine:
      leftStateMachine.addState(leftSupportState);
      leftStateMachine.addState(leftToeOffState);
      leftStateMachine.addState(leftSwingState);
      leftStateMachine.addState(leftStraightenState);


      // Assemble the Right State Machine:
      rightStateMachine.addState(rightSupportState);
      rightStateMachine.addState(rightToeOffState);
      rightStateMachine.addState(rightSwingState);
      rightStateMachine.addState(rightStraightenState);

      // Set the Initial States:

      leftStateMachine.setCurrentState(States.STRAIGHTEN);
      rightStateMachine.setCurrentState(States.SUPPORT);
   }


   private void balistic_walking_state_machine()
   {
      // Robot happens to walk in negative x direction.  Set vel positive just
      // so it makes intuitive sense.

      vel.set(-robot.qd_x.getDoubleValue());

      // Calculate forces on the feet

      left_force.set(robot.gc_lheel_fz.getDoubleValue() + robot.gc_ltoe_fz.getDoubleValue());
      if (left_force.getDoubleValue() > 5.0)
         left_cop.set(robot.gc_ltoe_fz.getDoubleValue() / left_force.getDoubleValue());
      else
         left_cop.set(0.5);

      right_force.set(robot.gc_rheel_fz.getDoubleValue() + robot.gc_rtoe_fz.getDoubleValue());
      if (right_force.getDoubleValue() > 5.0)
         right_cop.set(robot.gc_rtoe_fz.getDoubleValue() / right_force.getDoubleValue());
      else
         right_cop.set(0.5);


      // Actions in Each State

      leftStateMachine.doAction();
      rightStateMachine.doAction();

      // Communication between legs for now...

      // Subtract swing hip torque from the stance leg so less disturbance
      // on the body

      if (leftStateMachine.isCurrentState(States.SWING))
      {
         act_rh.set(act_rh.getDoubleValue() - act_lh.getDoubleValue());
      }

      if (rightStateMachine.isCurrentState(States.SWING))
      {
         act_lh.set(act_lh.getDoubleValue() - act_rh.getDoubleValue());
      }

      // Transition Conditions:

      leftStateMachine.checkTransitionConditions();
      rightStateMachine.checkTransitionConditions();

      // Torques at the joints:

      robot.tau_lh.set(act_lh.getDoubleValue() + pas_lh.getDoubleValue());
      robot.tau_lk.set(act_lk.getDoubleValue() + pas_lk.getDoubleValue());
      robot.tau_la.set(act_la.getDoubleValue() + pas_la.getDoubleValue());

      robot.tau_rh.set(act_rh.getDoubleValue() + pas_rh.getDoubleValue());
      robot.tau_rk.set(act_rk.getDoubleValue() + pas_rk.getDoubleValue());
      robot.tau_ra.set(act_ra.getDoubleValue() + pas_ra.getDoubleValue());
   }

   public YoVariableRegistry getRegistry()
   {
      return registry;
   }

   public class LeftHealUnloadedCondition implements StateTransitionCondition
   {
      public boolean checkCondition()
      {
         return (robot.gc_lheel_fz.getDoubleValue() < force_thresh.getDoubleValue())
                && ((robot.qd_x.getDoubleValue() < 0.0) && (robot.gc_lheel_x.getDoubleValue() > robot.q_x.getDoubleValue()))
                && ((robot.gc_lheel_x.getDoubleValue() > robot.gc_rheel_x.getDoubleValue()));
      }
   }


   public class RightHealUnloadedCondition implements StateTransitionCondition
   {
      public boolean checkCondition()
      {
         return (robot.gc_rheel_fz.getDoubleValue() < force_thresh.getDoubleValue())
                && ((robot.qd_x.getDoubleValue() < 0.0) && (robot.gc_rheel_x.getDoubleValue() > robot.q_x.getDoubleValue()))
                && ((robot.gc_rheel_x.getDoubleValue() > robot.gc_lheel_x.getDoubleValue()));
      }
   }


   public class LeftFootUnloadedCondition implements StateTransitionCondition
   {
      public boolean checkCondition()
      {
         return (left_force.getDoubleValue() < force_thresh.getDoubleValue());
      }
   }


   public class RightFootUnloadedCondition implements StateTransitionCondition
   {
      public boolean checkCondition()
      {
         return (right_force.getDoubleValue() < force_thresh.getDoubleValue());
      }
   }


   public class LeftFootTouchedDownCondition implements StateTransitionCondition
   {
      public boolean checkCondition()
      {
         return (robot.gc_ltoe_fs.getDoubleValue() == 1.0) || (robot.gc_lheel_fs.getDoubleValue() == 1.0);
      }
   }


   public class RightFootTouchedDownCondition implements StateTransitionCondition
   {
      public boolean checkCondition()
      {
         return (robot.gc_rtoe_fs.getDoubleValue() == 1.0) || (robot.gc_rheel_fs.getDoubleValue() == 1.0);
      }
   }


   public String getName()
   {
      return name;
   }
   
   public void initialize()
   {      
   }

   public String getDescription()
   {
      return getName();
   }

}
